"""
Created for obtaining the U-parameter phase based on correlation entropy, including
two-point correlators, four-point correlators for the prediction of the neural-network model

@author: Faluke Aikebaier
"""

### 1. libraries
from dmrgpy import fermionchain
import numpy as np
import cmath as cm
from numpy import linalg
from numpy.random import seed
import random

### 2. Functions
def get_fc(t,U,V,t1,h,mu,lamb):
    '''This function returns the many-body Hamiltonian'''
    fc = fermionchain.Spinful_Fermionic_Chain(n)
    H = 0
    
    # Hopping terms
    for i in range(n-1):
        H = H - t*fc.Cdagup[i]*fc.Cup[i+1]
        H = H - t*fc.Cdagdn[i]*fc.Cdn[i+1]
    for i in range(n-2): # hopping
        H = H + t1*fc.Cdagup[i]*fc.Cup[i+2]
        H = H + t1*fc.Cdagdn[i]*fc.Cdn[i+2]

    H = H + H.get_dagger() # Make Hermitian
    
    # Spin-orbit Coupling term
    for i in range(n-1):
        H = H + lamb*fc.Cdagup[i]*fc.Cdn[i+1]
        H = H - lamb*fc.Cdagdn[i]*fc.Cup[i+1]
    for i in range(1,n):
        H = H - lamb*fc.Cdagup[i]*fc.Cdn[i-1]
        H = H + lamb*fc.Cdagdn[i]*fc.Cup[i-1]
    
    # Diagonal terms
    for i in range(n):
        H = H + U*(fc.Nup[i]-0.5)*(fc.Ndn[i]-0.5)
        H = H + h*(fc.Nup[i]-fc.Ndn[i])
        H = H - mu*(fc.Nup[i]+fc.Ndn[i])
    
    for i in range(n-1):
        H = H + V*(fc.Nup[i]+fc.Ndn[i]-1)*(fc.Nup[i+1]+fc.Ndn[i+1]-1)
    
    fc.set_hamiltonian(H) # initialize the Hamiltonian
    
    fc.maxm = 80
    fc.nsweeps = 80
    
    return fc

def get_correlator_full(fc,wf,ii,jj,spin1,spin2):
    """This function returns two-point correlators"""
    if spin1=='up' and spin2=='up':
        corr=fc.Cdagup[ii].multiply_MO(fc.Cup[jj])
    if spin1=='up' and spin2=='dn':
        corr=fc.Cdagup[ii].multiply_MO(fc.Cdn[jj])
    if spin1=='dn' and spin2=='up':
        corr=fc.Cdagdn[ii].multiply_MO(fc.Cup[jj])
    if spin1=='dn' and spin2=='dn':
        corr=fc.Cdagdn[ii].multiply_MO(fc.Cdn[jj])

    return wf.aMb(corr,wf)

def get_correlator4_full(fc,wf,ii,jj,spin1,spin2):
    """This function returns four-point correlators"""
    if spin1=='up' and spin2=='up':
        if ii==jj:
            corr=0
        else:
            corr=fc.Nup[ii].multiply_MO(fc.Nup[jj])
    if spin1=='up' and spin2=='dn':
        corr=fc.Nup[ii].multiply_MO(fc.Ndn[jj])
    if spin1=='dn' and spin2=='up':
        corr=fc.Ndn[ii].multiply_MO(fc.Nup[jj])
    if spin1=='dn' and spin2=='dn':
        if ii==jj:
            corr=0
        else:
            corr=fc.Ndn[ii].multiply_MO(fc.Ndn[jj])

    return wf.aMb(corr,wf)

def correlation_matrix_full(n,fc,wf):
    """This function returns two-point correlation matrix"""
    corr_matrix=np.zeros((2*n,2*n))
    for ii in range(0,n):
        for jj in range(0,n):
            for kk in range(0,2):
                for ll in range(0,2):
                    corr_matrix[2*ii+kk,2*jj+ll]=get_correlator_full(fc,wf,ii,jj,spin_set[kk],spin_set[ll])
    return corr_matrix

def correlation4_matrix_full(n,fc,wf):
    """This function returns four-point correlation matrix"""
    corr_matrix=np.zeros((2*n,2*n))
    for ii in range(0,n):
        for jj in range(0,n):
            for kk in range(0,2):
                for ll in range(0,2):
                    corr_matrix[2*ii+kk,2*jj+ll]=get_correlator4_full(fc,wf,ii,jj,spin_set[kk],spin_set[ll])
    return corr_matrix

def correlation_entropy_full(corr_matrix):
    """This function returns the correlation entropy"""
    las2 = linalg.eigvalsh(corr_matrix) 
    Sc0=0
    for k in range(0,len(las2)):
        if las2[k]==0:
            Sc0+=0
        else:
            Sc0+=-las2[k]*cm.log(las2[k])
    return Sc0

### 3. Parameters and containers
#### 3.1 Basic parameters
n = 24 # number of sites of the Hubbard model
t=1 # nearest-neighbour hopping, also the unit energy of the system
spin_set=['up','dn'] # Represents two spin species, used in obtaining four-point correlators
#### 3.2 Containers for correlation entropy
Sc_list=[] # to be used for appending correlation entropy
Sc_list2=[]# to be used for appending normalized correlation entropy
Ulist=[] # to be used for appending values of U
lambdalist=[] # to be used for appending values of lambda
#### 3.3 Containers for two-point correlators
correlatror_list1, correlatror_list2, correlatror_list3, correlatror_list4=[], [], [], []
correlatror_list5, correlatror_list6, correlatror_list7, correlatror_list8=[], [], [], []
correlatror_list9, correlatror_list10, correlatror_list11, correlatror_list12=[], [], [], []
correlatror_list13, correlatror_list14, correlatror_list15, correlatror_list16=[], [], [], []
correlatror_list17, correlatror_list18, correlatror_list19, correlatror_list20=[], [], [], []
correlatror_list21, correlatror_list22, correlatror_list23, correlatror_list24=[], [], [], []
correlatror_list25, correlatror_list26, correlatror_list27, correlatror_list28=[], [], [], []
correlatror_list29, correlatror_list30, correlatror_list31, correlatror_list32=[], [], [], []
#### 3.4 Containers for four-point correlators
correlatror4_list1, correlatror4_list2, correlatror4_list3, correlatror4_list4=[], [], [], []
correlatror4_list5, correlatror4_list6, correlatror4_list7, correlatror4_list8=[], [], [], []
correlatror4_list9, correlatror4_list10, correlatror4_list11, correlatror4_list12=[], [], [], []
correlatror4_list13, correlatror4_list14, correlatror4_list15, correlatror4_list16=[], [], [], []
correlatror4_list17, correlatror4_list18, correlatror4_list19, correlatror4_list20=[], [], [], []
correlatror4_list21, correlatror4_list22, correlatror4_list23, correlatror4_list24=[], [], [], []
correlatror4_list25, correlatror4_list26, correlatror4_list27, correlatror4_list28=[], [], [], []
correlatror4_list29, correlatror4_list30, correlatror4_list31, correlatror4_list32=[], [], [], []

### 4. Obtaining data
NN=100 # Number of examples
for zz in range(0,NN):
    
    # Generating random variables
    seed(1)
    lamb=random.sample(list(np.arange(0,1.5,0.005)), 1)[0]
    mu=0*random.sample(list(np.arange(-2,2,0.001)), 1)[0]
    h=0*random.sample(list(np.arange(-2,2,0.001)), 1)[0]
    t1=0*random.sample(list(np.arange(0,0.9,0.005)), 1)[0]
    U=random.sample(list(np.arange(-10,10,0.005)), 1)[0]
    V=0*random.sample(list(np.arange(-8,8,0.005)), 1)[0]
    
    Ulist.append(U)
    lambdalist.append(lamb)
    
    # Many-body states and correlation entropy
    fc=get_fc(t,U,V,t1,h,mu,lamb)
    wf = fc.get_gs("DMRG")
    
    corr_mat=correlation_matrix_full(n,fc,wf)
    Sc=correlation_entropy_full(corr_mat)
    
    corr4_mat=correlation4_matrix_full(n,fc,wf)
    
    # --- Two - point Correlators ---
    correlatror_list1.append(corr_mat[n-3][n-3])
    correlatror_list2.append(corr_mat[n-3][n-2])
    correlatror_list3.append(corr_mat[n-3][n-1])
    correlatror_list4.append(corr_mat[n-3][n])
    
    correlatror_list5.append(corr_mat[n-2][n-2])
    correlatror_list6.append(corr_mat[n-2][n-1])
    correlatror_list7.append(corr_mat[n-2][n])
    correlatror_list8.append(corr_mat[n-2][n+1])
    
    correlatror_list9.append(corr_mat[n-1][n-1])
    correlatror_list10.append(corr_mat[n-1][n])
    correlatror_list11.append(corr_mat[n-1][n+1])
    correlatror_list12.append(corr_mat[n-1][n+2])
    
    correlatror_list13.append(corr_mat[n][n])
    correlatror_list14.append(corr_mat[n][n+1])
    correlatror_list15.append(corr_mat[n][n+2])
    correlatror_list16.append(corr_mat[n][n+3])
    
    correlatror_list17.append(corr_mat[n+1][n+1])
    correlatror_list18.append(corr_mat[n+1][n+2])
    correlatror_list19.append(corr_mat[n+1][n+3])
    correlatror_list20.append(corr_mat[n+1][n+4])
    
    correlatror_list21.append(corr_mat[n+2][n+2])
    correlatror_list22.append(corr_mat[n+2][n+3])
    correlatror_list23.append(corr_mat[n+2][n+4])
    correlatror_list24.append(corr_mat[n+2][n+5])
    
    correlatror_list25.append(corr_mat[n+3][n+3])
    correlatror_list26.append(corr_mat[n+3][n+4])
    correlatror_list27.append(corr_mat[n+3][n+5])
    correlatror_list28.append(corr_mat[n+3][n+6])
    
    correlatror_list29.append(corr_mat[n+4][n+4])
    correlatror_list30.append(corr_mat[n+4][n+5])
    correlatror_list31.append(corr_mat[n+4][n+6])
    correlatror_list32.append(corr_mat[n+4][n+7])
    
    # --- Four - point Correlators ---
    correlatror4_list1.append(corr4_mat[n-3][n-2])
    correlatror4_list2.append(corr4_mat[n-3][n-1])
    correlatror4_list3.append(corr4_mat[n-3][n])
    correlatror4_list4.append(corr4_mat[n-3][n+1])
    
    correlatror4_list5.append(corr4_mat[n-2][n-1])
    correlatror4_list6.append(corr4_mat[n-2][n])
    correlatror4_list7.append(corr4_mat[n-2][n+1])
    correlatror4_list8.append(corr4_mat[n-2][n+2])
    
    correlatror4_list9.append(corr4_mat[n-1][n])
    correlatror4_list10.append(corr4_mat[n-1][n+1])
    correlatror4_list11.append(corr4_mat[n-1][n+2])
    correlatror4_list12.append(corr4_mat[n-1][n+3])
    
    correlatror4_list13.append(corr4_mat[n][n+1])
    correlatror4_list14.append(corr4_mat[n][n+2])
    correlatror4_list15.append(corr4_mat[n][n+3])
    correlatror4_list16.append(corr4_mat[n][n+4])
    
    correlatror4_list17.append(corr4_mat[n+1][n+2])
    correlatror4_list18.append(corr4_mat[n+1][n+3])
    correlatror4_list19.append(corr4_mat[n+1][n+4])
    correlatror4_list20.append(corr4_mat[n+1][n+5])
    
    correlatror4_list21.append(corr4_mat[n+2][n+3])
    correlatror4_list22.append(corr4_mat[n+2][n+4])
    correlatror4_list23.append(corr4_mat[n+2][n+5])
    correlatror4_list24.append(corr4_mat[n+2][n+6])
    
    correlatror4_list25.append(corr4_mat[n+3][n+4])
    correlatror4_list26.append(corr4_mat[n+3][n+5])
    correlatror4_list27.append(corr4_mat[n+3][n+6])
    correlatror4_list28.append(corr4_mat[n+3][n+7])
    
    correlatror4_list29.append(corr4_mat[n+4][n+5])
    correlatror4_list30.append(corr4_mat[n+4][n+6])
    correlatror4_list31.append(corr4_mat[n+4][n+7])
    correlatror4_list32.append(corr4_mat[n+4][n+8])
    
    # Correlation entropy
    Sc_list.append(Sc.real)
    Sc_list2.append((Sc.real)/(n))

### 5. Storing data
data_X=np.zeros([len(Sc_list2),64])
data_Y=np.zeros([len(Sc_list2),1])
for n in range(len(correlatror_list1)):
    data_X[n,0]=correlatror_list1[n]
    data_X[n,1]=correlatror_list2[n]
    data_X[n,2]=correlatror_list3[n]
    data_X[n,3]=correlatror_list4[n]
    
    data_X[n,4]=correlatror_list5[n]
    data_X[n,5]=correlatror_list6[n]
    data_X[n,6]=correlatror_list7[n]
    data_X[n,7]=correlatror_list8[n]
    
    data_X[n,8]=correlatror_list9[n]
    data_X[n,9]=correlatror_list10[n]
    data_X[n,10]=correlatror_list11[n]
    data_X[n,11]=correlatror_list12[n]
    
    data_X[n,12]=correlatror_list13[n]
    data_X[n,13]=correlatror_list14[n]
    data_X[n,14]=correlatror_list15[n]
    data_X[n,15]=correlatror_list16[n]
    
    data_X[n,16]=correlatror_list17[n]
    data_X[n,17]=correlatror_list18[n]
    data_X[n,18]=correlatror_list19[n]
    data_X[n,19]=correlatror_list20[n]
    
    data_X[n,20]=correlatror_list21[n]
    data_X[n,21]=correlatror_list22[n]
    data_X[n,22]=correlatror_list23[n]
    data_X[n,23]=correlatror_list24[n]
    
    data_X[n,24]=correlatror_list25[n]
    data_X[n,25]=correlatror_list26[n]
    data_X[n,26]=correlatror_list27[n]
    data_X[n,27]=correlatror_list28[n]
    
    data_X[n,28]=correlatror_list29[n]
    data_X[n,29]=correlatror_list30[n]
    data_X[n,30]=correlatror_list31[n]
    data_X[n,31]=correlatror_list32[n]
    
    data_X[n,32]=correlatror4_list1[n]
    data_X[n,33]=correlatror4_list2[n]
    data_X[n,34]=correlatror4_list3[n]
    data_X[n,35]=correlatror4_list4[n]
    
    data_X[n,36]=correlatror4_list5[n]
    data_X[n,37]=correlatror4_list6[n]
    data_X[n,38]=correlatror4_list7[n]
    data_X[n,39]=correlatror4_list8[n]
    
    data_X[n,40]=correlatror4_list9[n]
    data_X[n,41]=correlatror4_list10[n]
    data_X[n,42]=correlatror4_list11[n]
    data_X[n,43]=correlatror4_list12[n]
    
    data_X[n,44]=correlatror4_list13[n]
    data_X[n,45]=correlatror4_list14[n]
    data_X[n,46]=correlatror4_list15[n]
    data_X[n,47]=correlatror4_list16[n]
    
    data_X[n,48]=correlatror4_list17[n]
    data_X[n,49]=correlatror4_list18[n]
    data_X[n,50]=correlatror4_list19[n]
    data_X[n,51]=correlatror4_list20[n]
    
    data_X[n,52]=correlatror4_list21[n]
    data_X[n,53]=correlatror4_list22[n]
    data_X[n,54]=correlatror4_list23[n]
    data_X[n,55]=correlatror4_list24[n]
    
    data_X[n,56]=correlatror4_list25[n]
    data_X[n,57]=correlatror4_list26[n]
    data_X[n,58]=correlatror4_list27[n]
    data_X[n,59]=correlatror4_list28[n]
    
    data_X[n,60]=correlatror4_list29[n]
    data_X[n,61]=correlatror4_list30[n]
    data_X[n,62]=correlatror4_list31[n]
    data_X[n,63]=correlatror4_list32[n]
    
    data_Y[n]=Sc_list2[n]

### 6. Storing U and V
data_Ulambda=np.zeros([len(Sc_list2),2])
for n in range(len(correlatror_list1)):
    data_Ulambda[n,0]=Ulist[n]
    data_Ulambda[n,1]=lambdalist[n]
    
### 7. Saving data
np.save("data_X_n24_j.npy", data_X)
np.save("data_Y_n24_j.npy", data_Y)
np.save("data_Ulambda_24_j.npy", data_Ulambda)
